/***************************************************************************
 * CVSID: $Id$
 *
 * probe-input.c : Probe input devices
 *
 * Copyright (C) 2004 David Zeuthen, <david@fubar.dk>
 *
 * Licensed under the Academic Free License version 2.1
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 **************************************************************************/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <errno.h>
#include <fcntl.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#ifdef HAL_LINUX_INPUT_HEADER_H
  #include HAL_LINUX_INPUT_HEADER_H
#else
  #include <linux/input.h>
#endif

#include "libhal/libhal.h"
#include "../../logger.h"

/* we must use this kernel-compatible implementation */
#define BITS_PER_LONG (sizeof(long) * 8)
#define NBITS(x) ((((x)-1)/BITS_PER_LONG)+1)
#define OFF(x)  ((x)%BITS_PER_LONG)
#define BIT(x)  (1UL<<OFF(x))
#define LONG(x) ((x)/BITS_PER_LONG)
#define test_bit(bit, array)    ((array[LONG(bit)] >> OFF(bit)) & 1)

static const int sw_bits_headphone_insert[] = {
	SW_HEADPHONE_INSERT,
	SW_MICROPHONE_INSERT,
	SW_LINEOUT_INSERT,
	SW_VIDEOOUT_INSERT,
	SW_JACK_PHYSICAL_INSERT,
	-1
};
static const int sw_bits_lid[] = { 
	SW_LID, 
	-1 
};
static const int sw_bits_tablet_mode[] = {
	SW_TABLET_MODE, 
	-1
};
static const int sw_bits_keypad_slide[] = {
	SW_KEYPAD_SLIDE, 
	-1
};
static const int sw_bits_camera_lens_cover[] = {
	SW_CAMERA_LENS_COVER, 
	-1
};
#ifdef SW_RADIO
static const int sw_bits_radio[] = {
	SW_RADIO,
	-1
};
#endif

static const struct {
	const char *name;
	const int  *bits;
} lut[] = {
	{"headphone_insert",  sw_bits_headphone_insert },
	{"jack_insert",       sw_bits_headphone_insert },

	{"lid",               sw_bits_lid },
	{"tablet_mode",       sw_bits_tablet_mode },
	{"keypad_slide",      sw_bits_keypad_slide },
	{"camera_lens_cover", sw_bits_camera_lens_cover },

#ifdef SW_RADIO
	{"radio",             sw_bits_radio },
#endif
	{ NULL, NULL }
};

int 
main (int argc, char *argv[])
{
	/* If the probe process exits with error, hal daemon will
	 * remove the input device -> this should be avoided if
	 * possible */
	int ret = EXIT_SUCCESS;

	int fd = -1;
	LibHalContext *ctx = NULL;
	DBusError error = DBUS_ERROR_INIT;

	const char *udi = NULL;
	const char *device_file = NULL;
	const char *button_type = NULL;
	const int *sw_bits = NULL;
	dbus_bool_t value = FALSE;
	long bitmask[NBITS(SW_MAX)];
	size_t i;

	setup_logger ();
	dbus_error_init (&error);

	if( !(udi = getenv ("UDI")) ) {
		HAL_WARNING (("$%s is not set", "UDI"));
		goto out;
	}
	if( !(device_file = getenv ("HAL_PROP_INPUT_DEVICE")) ) {
		HAL_WARNING (("$%s is not set", "HAL_PROP_INPUT_DEVICE"));
		goto out;
	}
	if( !(button_type = getenv ("HAL_PROP_BUTTON_TYPE")) ) {
		HAL_WARNING (("$%s is not set", "HAL_PROP_BUTTON_TYPE"));
		goto out;
	}

	HAL_DEBUG (("Doing probe-input for %s (type=%s, udi=%s)",
		    device_file, button_type, udi));

	for( i = 0; lut[i].name; ++i )
		if( !strcmp(lut[i].name, button_type) ) {
			sw_bits = lut[i].bits;
			break;
		}
	if( !sw_bits ) {
		HAL_WARNING (("Unknown $%s: %s", "HAL_PROP_BUTTON_TYPE",
			      button_type));
		goto out;
	}
	
	if ((ctx = libhal_ctx_init_direct (&error)) == NULL) {
		HAL_ERROR (("Cannot initialize hal context"));
		ret = EXIT_FAILURE;
		goto out;
	}

	if( (fd = open (device_file, O_RDONLY)) < 0 ) {
		HAL_WARNING (("Cannot open %s: %s", device_file,
			    strerror (errno)));
		goto out;
	}

	memset(bitmask, 0, sizeof bitmask);
	if (ioctl (fd, EVIOCGSW(SW_MAX), bitmask) < 0) {
		HAL_WARNING (("ioctl EVIOCGSW failed"));
		goto out;
	}

	for (i = 0; sw_bits[i] != -1; ++i )
		if (test_bit (sw_bits[i], bitmask)) {
			value = TRUE;
			break;
		}

	if( !libhal_device_set_property_bool (ctx, udi, "button.state.value",
					      value, &error) ) {
	  HAL_WARNING (("Cannot set property %s: %s = %s", device_file,
		      "button.state.value", value ? "True" : "False"));
	  goto out;
	}

out:
	if (fd >= 0)
		close (fd);

	LIBHAL_FREE_DBUS_ERROR (&error);

	if (ctx != NULL) {
		libhal_ctx_shutdown (ctx, &error);
		LIBHAL_FREE_DBUS_ERROR (&error);
		libhal_ctx_free (ctx);
	}

	return ret;
}
